@@ -1,7 +1,7 @@
 echo "Point 1"
 exec 3</etc/passwd
-exec 4>a
-exec 5>b
+exec 4>/tmp/a
+exec 5>/tmp/b
 echo "Point 2"
 echo to a 1>&4
 echo to b 1>&5
