@@ -18,8 +18,8 @@
    with Bash; see the file COPYING.  If not, write to the Free Software
    Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */
 
-#if !defined (__TRAP_H__)
-#define __TRAP_H__
+#if !defined (_TRAP_H_)
+#define _TRAP_H_
 
 #include "stdc.h"
 
@@ -36,19 +36,27 @@
 #define DEFAULT_SIG SIG_DFL
 #define IGNORE_SIG  SIG_IGN
 
+/* Special shell trap names. */
+#define DEBUG_TRAP NSIG
+#define EXIT_TRAP 0
+
 #define signal_object_p(x) (decode_signal (x) != NO_SIG)
 
-extern char *trap_list[NSIG];
+extern char *trap_list[];
 
 /* Externally-visible functions declared in trap.c. */
 extern void initialize_traps __P((void));
 extern void run_pending_traps __P((void));
 extern void maybe_set_sigchld_trap __P((char *));
 extern void set_sigchld_trap __P((char *));
+extern void set_debug_trap __P((char *));
+extern void set_sigint_trap __P((char *));
 extern void set_signal __P((int, char *));
 extern void restore_default_signal __P((int));
 extern void ignore_signal __P((int));
 extern int run_exit_trap __P((void));
+extern void run_trap_cleanup __P((int));
+extern void run_debug_trap __P((void));
 extern void free_trap_strings __P((void));
 extern void reset_signal_handlers __P((void));
 extern void restore_original_signals __P((void));
@@ -59,7 +67,8 @@ extern int decode_signal __P((char *));
 extern void run_interrupt_trap __P((void));
 extern int maybe_call_trap_handler __P((int));
 extern int signal_is_trapped __P((int));
-extern int signal_is_special __P((int));
 extern int signal_is_ignored __P((int));
+extern int signal_is_special __P((int));
+extern void set_signal_ignored __P((int));
 
-#endif /* __TRAP_H__ */
+#endif /* _TRAP_H_ */
